'' Basic I2C Routines  Version 2.1
'' (c) 2014 Tracy Allen, EME Systems LLC
'' Terms of MIT liscense, see end of object code
'' Permission is given to use this in any program for the Parallax
'' Propeller processor as long as this copyright notice is included.
'' About this implementation:
''    -- Init(scl,sda) can use any combination of pins for scl,sda
''    -- both scl and sda need pullups.  The routines manipulate dira[.] only.  Won't work with eeprom on some Parallax development boards.
''    -- Use 8-bit device addresses, for devSel.  That is the 7-bit address shifted left by 1.
''    -- negx is used as a flag to skip the address bytes, for devices that use only a device ID, no address
''    -- To handle 16 bit addresses, pass the address in as a negative value.  A positive address value signifies 8 bit address
''    -- This object does not directly support addresses greater than 16 bits.  That must be handled in application code by patching the device ID.
''    -- The count parameter for strings and numbers can be passed as a negative number. That converts smallEndian<-->largeEndian on the fly.
''    -- Methods for reading and writing numbers (1, 2, 3 or 4 bytes) can likewise do msbyte<-->lsbyte inversions
''    -- Supports clock streching on any rising edge of scl, with timeout parameter and method to change the timeout parameter
''    -- sda and scl parameters are stored as variables passed via init(..)
''       - each object that declares i2c_1bus must call the init(.,.) method.
''       - different objects can init this object with the same pins without conflict
''       - Each object instance declared uses another pair of bytes, but there is still only one copy of the object code
''       - take care only one device accesses the bus at a time; No locks are implemented!!!
''       - individual method calls do not reference the pins, those are remembered in the object instance variables.
''    -- object can be used for multiple busses, init the object with its own set of pins


CON
   ACK      = 0                        ' I2C Acknowledge
   NAK      = 1                        ' I2C No Acknowledge
   DRIVELOW = 1                        ' state of dira[.] is output to drive pin low
   PULLHIGH = 0                        ' state of dira[.] is input to allow pin to be pulled high
   LOW = 0                             ' the output register of both scl and sda will always be low, control is via dira[.] only
   NO_ADDR   = negx                     ' as a flag in the addrReg parameter position in the method call, causes method to skip sending address byte(s)
   NEXTUP = negx                        ' ditto
   DEFAULT_HOLD_TICKS = 8_000_000          ' in seconds for clock stretching, 8M = 0.1 second for clock hold

   BOOTPIN  = 28                       ' I2C Boot EEPROM SCL Pin
   EEPROM   = $A0                      ' I2C EEPROM Device 8-bit devsel (use negative of address value for 16 bit address parameter)

OBJ

VAR
  byte scl, sda
  long  maxHoldTicks  ' 8_000_000 is 0.1 second at 80 MHz, for clock stretching


CON ' iniitialization methods
PUB Init28
  return Init(28, 29)

PUB Init(_scl,_sda)                               ' pass in both pins, call to set up the buss or to change from one buss to another
  sda := _sda
  scl := _scl
  maxHoldTicks := DEFAULT_HOLD_TICKS                     ' default clock hold, can set with method below
  dira[scl] := PULLHIGH                                 ' both scl and sda are pulled high by resistor
  dira[sda] := PULLHIGH
  outa[scl] := LOW                                      ' leave a zero in the scl and sda output latches.
  outa[sda] := LOW                                      ' so turning pin to output will drive the pin low.
  return  ina[scl] << 1 | ina[sda]                         ' should return 3 when pullups are present, bus is ready

PUB Twiddle                                             ' some devices can use this to get out of a bad state.
  repeat 9
    dira[scl] := DRIVELOW                             ' Put out up to 9 clock pulses
    dira[scl] := PULLHIGH
      if ina[scl] == 0
        Hold
      if ina[sda]                                       ' Repeat if SDA not driven high
        quit
  return ina[scl] << 1 & ina[sda]                         ' should return 3 when pullups are present, bus is ready

CON ' low level methods
PUB Start                                               ' SDA goes HIGH to LOW with SCL HIGH
  dira[sda] := PULLHIGH
  dira[scl] := PULLHIGH
  if ina[scl] == 0
      Hold
  dira[sda] := DRIVELOW
  dira[scl] := DRIVELOW


PUB Stop                                                ' SDA goes LOW to HIGH with SCL High
  dira[sda] := DRIVELOW
  dira[scl] := PULLHIGH
  if ina[scl] == 0
      Hold
  dira[sda] := PULLHIGH

PUB Write(data) : ackbit
'' Write i2c data.  Data byte is output MSB first, SDA data line is valid
'' only while the SCL line is HIGH.  SDA can change only when SCL is low.
'' Data is always 8 bits (+ ACK/NAK).
'' SCL is left in the LOW state, sda state is indeterminant
'' returns 0 if slave ACK, 1 if NAK
  ackbit := 0
  data := !data << 24
  repeat 8
    dira[sda] := data <-= 1
    dira[scl] := PULLHIGH
    if ina[scl] == 0
      Hold
    dira[scl] := DRIVELOW
  dira[sda] := PULLHIGH         ' allow slave to control sda
  dira[scl] := PULLHIGH         ' this edge latches the ack bit from slave
  if ina[scl] == 0              ' allow slave to hold
    Hold
  ackbit := ina[sda]
  dira[scl] := DRIVELOW         ' sda can change now

PUB Read(ackbit): data
'' Read in i2c data, Data byte is input MSB first, SDA data line is
'' valid only while the SCL line is HIGH, and changes only when SCL is low
'' SCL left in LOW state, sda state is indeterminant
'' returns byte value
  dira[sda] := PULLHIGH
  repeat 8
    dira[scl] := PULLHIGH
    if ina[scl] == 0
      Hold
    data := (data << 1) | ina[sda]
    dira[scl] := DRIVELOW
  dira[sda] := !ackbit
  dira[scl] := PULLHIGH         ' slave latches the ACK or NAK
  if ina[scl] == 0              ' allow slave to hold
    Hold
  dira[scl] := DRIVELOW    ' sda state is indeterminant, device can change it when scl is low

PRI Hold | ticks
' this implements clock hold, with a timeout
  ticks := cnt + maxHoldTicks
  repeat until (ina[scl]==1) or (cnt - ticks > 0)

PUB SetMaxHoldTime(milliseconds)
  return maxHoldTicks := clkfreq/1_000 * (milliseconds)

PUB Hello(devSel) : ackbit
'' The device will not respond if it is busy, or if it is not present on the i2c buss
'' function call returns 0 if device at devSel responded (ACK, present and ready), otherwise 1 (NAK, not present, or currently busy)
'' note this can return fake ack if the bus is locked low, need also to test for pullups.
   Start
     ackbit := Write(devSel)
   Stop

CON ' high level Read methods
PUB ReadPage(devSel, address, dataPtr, byteCount) : ackbit | step1
'' Read in a counted block of i2c data.  Device select code is devSel, 8 bit form with lsb=0
'  Device starting address is "address". Prop address is dataPtr.  Number of bytes is count.
'' address is normally positive 8 bit address; Make it negative to force a two-byte address (EEPROM)
'' For eeproms > 64 kbytes, must handle 17th and higher bits ad hoc to merge with devSel
'' Use address=negx as a flag to skip the device address and go right into the bytes (some devices, MAX1161x, devsel only).
'' The data returned from the device is placed in the Prop byte array pointed to at entry by dataPtr.
'' Normally byteCount is positive and bytes are the same order in Prop as they are in the device.
'' Make count negative to force reverse order, endedness swap.
'' The function call returns 0 if all acks are okay, 1 otherwise, The ACKS/NAKS on read come only from writing devSel and addrReg
'' ----------
  step1 := 1
  if address > negx
    Start                       '
    ackbit := Write(devSel)
    if address < 0             ' use negative addrReg value to invoke 16 bit address
      || address
      ackbit |= Write(address >> 8)              ' msb
    ackbit |= Write(address)                    ' lsb
  Start                          ' Reselect the device for reading
  ackbit |= Write(devSel | 1)  ' 8 bit address, 1 in lsb for read operation
  if byteCount < 0                  ' edit1, consolidated +/- with step1 variable
    -step1
    ||byteCount
    dataPtr += byteCount - 1
  repeat byteCount - 1
    byte[dataPtr] := Read(ACK)   ' same order in both prop and device
    dataPtr += step1           ' step1= +/- 1 for order of bytes moved from device to prop
  byte[dataPtr] := Read(NAK)
  Stop
  ' return 0 if all ACKs okay, 1 if one or more NAKs in the setup

PUB ReadByte(devSel, address)
'' reads one byte from device.
'' positive value in address sends 8 bit address, negative value in address sends 16 bit device address
'' negx in address skips sending any address bytes  (devSel only, followed by repeated start and data read)
   ReadPage(devSel, address, @result, 1)

PUB ReadRegister(devSel)
'' reads one byte from a simple device that needs only device selection.
  Start
  Write(devSel | 1)
  result := Read(NAK)
  Stop

PUB ReadNumber(devSel, address, byteCount)
'' count is length of the number in bytes (1,2,3,4), to be transferred from device to Prop.
'' positive count transfers number byte for byte, negative count swaps the endedness, that is, msb from device goes to lsb in Prop.
'' positive value in address sends 8 bit address, negative value in address sends 16 bit address
'' negx in address skips sending any address bytes  (devSel only)
  ReadPage(devSel, address, @result, byteCount)


CON ' high level Write methods
PUB WritePage(devSel, address, dataPtr, byteCount) : ackbit | step1
'' Write out a block of i2c data.  Device select code is devSel, 8 bit form.
'  Device starting address is address. Data address is dataPtr.  Number of bytes is count.
'' Use negative address to specify two byte address.
'' Use address=negx to specify null, no address bytes, goes right into the data bytes
'' Use count=0 to skip sending any data (i.e. devSel and addrReg only.) (Same as WriteRegister)
'' For eeproms > 64 kbytes, must handle 17th and higher bits ad hoc to merge with devSel
'' Return zero if all acks are okay, 1 otherwise
'' ----------
  step1 := 1
  Start
  ackbit := Write(devSel)   '  devSel is 8 bit, zero in lsb
  if address > negx         ' negx flag causes skip direct to sending bytes
    if address<0             ' negative addrReg signals 2 byte address (eeprom)
      ||address
      ackbit |= Write(address >> 8)   ' msb
    ackbit |= Write(address)           ' lsb, or the one and only address byte
  if byteCount<0                                            ' edit1, consolidated +/- with step1 variable
    -step1
    ||byteCount
    dataptr += byteCount - 1
  repeat byteCount
     ackbit |= Write(byte[dataPtr])
     dataPtr += step1                                  ' step1= +/- 1 for order of bytes moved from prop to device
  Stop
  ' return 0 if all ACKs okay, 1 if one or more NAKs

PUB WriteRegister(devSel, mybyte) : ackbit
'' Writes one byte to a device register.
'' This might be a very simple device, or the register might be a pointer or config register.
'' Returns 0 if all ACKs are okay, 1 if one or more NAKs
  Start
  ackbit := Write(devSel)   '  devSel is 8 bit, zero in lsb
  ackbit |= Write(mybyte)
  Stop


PUB WriteByte(devSel, address, myByte) : ackbit
'' writes one byte to device.  This is faster than WriteNumber for a single byte
'' positive value in address sends 8 bit address, negative value signals 16 bit address
'' if address is not needed, us WriteRegister instead of WriteByte
'' function call returns 0 if all ACKs are okay, 1 if one or more NAKs
  Start
  ackbit := Write(devSel)   '  devSel is 8 bit, zero in lsb
  if address < 0
    || address
    ackbit |= Write(address>>8)
  ackbit |= Write(address)
  ackbit |= Write(myByte)
  Stop

PUB WriteNumber(devSel, address, number, byteCount)
'' count is length of the number in bytes, 1--4, to be transferred from Prop to device.
'' count as a posibive number transfers number byte for byte, negative count swaps the endedness, that is, msb from prop goes to lsb in device.
'' positive value in address sends 8 bit address, negative value in address sends 16 bit address
'' negx in address skips sending any address bytes  (sends devSel only then the number)
'' function call returns 0 if all ACKs are okay, 1 if one or more NAKs
   return  WritePage(devSel, address, @number, byteCount)


CON

{{
┌──────────────────────────────────────────────────────────────────────────────────────┐
│                           TERMS OF USE: MIT License                                  │
├──────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this  │
│software and associated documentation files (the "Software"), to deal in the Software │
│without restriction, including without limitation the rights to use, copy, modify,    │
│merge, publish, distribute, sublicense, and/or sell copies of the Software, and to    │
│permit persons to whom the Software is furnished to do so, subject to the following   │
│conditions:                                                                           │
│                                                                                      │
│The above copyright notice and this permission notice shall be included in all copies │
│or substantial portions of the Software.                                              │
│                                                                                      │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,   │
│INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A         │
│PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT    │
│HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION     │
│OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE        │
│SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                                │
└──────────────────────────────────────────────────────────────────────────────────────┘
}}

