''=============================================================================
'' @file     sht2x_1bus.spin   version 1.0
'' @target   Propeller with Sensirion SHT2x
'' @author   Thomas Tracy Allen, EME Systems
'' Copyright (c) 2013 EME Systems LLC
'' See end of file for terms of use.
'' version 1.0
'' uses
''
''=============================================================================

CON

  DEV_ID = %1000_0000                                   ' only one device ID for SHT2x  8bit:$80 7bit:$40

  CMD_TEMP_HOLD = %1110_0011                             ' measure temperature using clock hold
  CMD_RH_HOLD    = %1110_0101                              ' measure humidity using clock hold

  CMD_TEMP_TRIGGER  = %1111_0011                             ' measure temperature, pollwait
  CMD_RH_TRIGGER  = %1111_0101                             ' measure temperature, pollwait

  CMD_READ_USER = %1110_0111                           ' also called user register
  CMD_WRITE_USER = %1110_0110                              ' write status (user register)

  CMD_RESET_I2C       = %1111_1110                              ' soft reset


'' user status register bitmapped, read before write
'' bits 7,0  resolution r/w
'' bit 6    battery level  r,  0 means >2.25, 1 means <2.25
'' bits 3-5   reserved, do not mess with these (read before write)
'' bit 2   heater, 0=disable, 1=enable
'' bit 1   autoreload settings, 1=disable, do not use
'' default is %00xxx010, 12 bit RH, 14 bit temp, heater off, autoreload disabled
  RES_12_14     = %0_000000_0     ' default, 12 bit humidity, 14 bit temperature, maximum hold 85ms
  RES_8_12      = %0_000000_1     ' maximum hold 22ms
  RES_10_13     = %1_000000_0      ' maximum hold 29 ms
  RES_11_11     = %1_000000_1     ' maximum hold 15 ms
  HEATER_ON     = %00000_1_00
  HEATER_OFF    = %00000_0_00

  HIGH = RES_12_14              ' these are constants that can be used with the SetResolution method
  LOW = RES_11_11               ' -- only implemented two of the possible resolution combinations.
  TC = CMD_TEMP_TRIGGER         ' these are constants that can be used with the TriggerRead method
  RH = CMD_RH_TRIGGER

OBJ
  i2c : "i2c_1bus"

VAR
  byte  which         ' for polling mode, remembers which mode was triggered, TC or RH,for subsequent calculations

PUB Init(scl, sda)
  result := (i2c.Init(SCL, SDA) << 1) | PollReady
  ' returns 6 if pullups good and device ACKed, anything else has a problem
  i2c.SetMaxHoldTime(90)  ' maximum conversion time using maximum resoltion is 85ms for 14 bit temperature

CON ' the following two methods use the clock hold option of the SHT2x series.
PUB readTemperature
  result := i2c.ReadNumber(DEV_ID, CMD_TEMP_HOLD, -2)  ' note -2 receives two bytes, msb first, swaps to return lsb first
  return Raw2Temperature(result)   ' returns value in 100ths xxx.xx degC

PUB readHumidity
  result := i2c.ReadNumber(DEV_ID, CMD_RH_HOLD, -2)
  return Raw2Humidity(result)   ' returns value in tenths, xx.x %RH

CON ' the following are for polled operation, alternative to the above. Trigger the read of TC or RH, then after delay or polling, retrieve the value
PUB TriggerRead(TC_RH)   ' use either sht#TC or sht#RH
  which := TC_RH
  i2c.WriteRegister(DEV_ID, TC_RH)

PUB PollReady
  return i2c.hello(DEV_ID)

PUB ReadResult
  result := i2c.ReadNumber(DEV_ID, negx, -2)    ' note: 2 bytes returned. endedness swap, negx means skip sending address psrameter.
  if which == TC
    return Raw2Temperature(result)
  elseif which == RH
    return Raw2Humidity(result)
  else
    return negx                                         ' this would mean a bad call

CON ' the following are utility methods
PUB Raw2Temperature(raw)
  return (raw * 17572) >> 16 - 4685   ' returns value in 100ths xxx.xx degC

PUB Raw2Humidity(raw)
  return (raw * 1250) >> 16 -  60

PUB SetResolution(high_low)  ' use sht#LOW or sht#HIGH (default). #LOW for fast 11 bit RH and degC, fast, or #HIGH for slow 14 bit degC 12 bit RH.
  WriteUser(high_low)

PUB Heater(on_off)  ' use sht#HEATER_ON or sht#HEATER_OFF (default)
  WriteUser(on_off)

PUB ReadUser
  return i2c.ReadByte(DEV_ID, CMD_READ_USER)

PUB WriteUser(value)
  ' have to read before write
  result := ReadUser
  result := (result & %0_1_111_010) | (value & 1_0_000_101)   ' keep reserved bits, allow change in resolution and heater only
  i2c.WriteByte(DEV_ID, CMD_WRITE_USER, result)
  ' e.. sht.WriteUser(sht#LO_RES | sht#HEATER_ON)


{{
┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │                                                            
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │ 
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
}}
